import os
import re
import sys

from glyles.glycans.factory import MonomerFactory
from glyles.grammar.parse import Glycan


def preprocess_glycans(glycan, glycan_list, glycan_file):
    """
    Preprocess the static inputs for the parsing into one single list

    Args:
        glycan (str): single glycan to parse
        glycan_list (List[str]): list of glycans to parse
        glycan_file (str): filepath of file to read glycans from

    Returns:
        list of glycans in the order they are handed to the function, i.e glycan, glycan_list, glycan_file
    """
    glycans = []

    # fill the list of glycans to convert
    if glycan is not None:
        glycans.append(glycan)
    if glycan_list is not None:
        glycans += glycan_list
    if glycan_file is not None:
        # check if the file is valid and read it out
        if not os.path.isfile(glycan_file):
            pass
        for line in open(glycan_file, "r").readlines():
            glycans.append(line.strip())
    return glycans


def parsable_glycan(glycan):
    """
    Check if the glycan string is parsable by removing all connections and brackets. Then remove all known monomers
    from the remaining string. What remains are the unknown monomers that cannot be parsed or an empty string if the
    glycan can be parsed.

    Args:
        glycan (str): IUPAC representation of a glycan

    Returns:
        True if the glycan can be parsed in the parser
    """
    glycan = re.sub("[\(].*?[\)]", "", glycan)
    glycan = glycan.replace("[", "").replace("]", "").upper()
    for monomer in MonomerFactory.monomers():
        glycan = glycan.replace(monomer, " ")
    return len(glycan.replace(" ", "")) == 0


def convert(glycan=None, glycan_list=None, glycan_file=None, glycan_generator=None, output_file=None, returning=False,
            silent=True):
    """
    General user interaction interface to use this library.

    Args:
        glycan (str): Single glycan to be converted from IUPAC to SMILES
        glycan_list (List[str]): list of glycans to convert
        glycan_file (str): File to read the glycans from
        glycan_generator (generator): generator yielding iupac representation.
            Together with output_generator=True this does not create any lists
        output_file (str): File to save the converted glycans in
        returning (bool): Flag indicating to return a list of tuples
        silent (bool): Flag indicating to have no prints from this method

    Returns:
        Nothing
    """

    glycans = preprocess_glycans(glycan, glycan_list, glycan_file)
    if len(glycans) == 0 and glycan_generator is None:
        if not silent:
            print("List of glycans is empty")
        return

    # determine the output format
    if output_file is not None:
        if os.path.isdir(os.path.dirname(os.path.abspath(output_file))):
            output = open(output_file, "w")
        else:
            if not silent:
                print("Path of output-file does not exist! Results will be printed on stdout.", file=sys.stderr)
            output = sys.stdout
    else:
        if returning:
            output = []
        else:
            if not silent:
                print("No output-file specified, results will be printed on stdout.")
            output = sys.stdout

    if len(glycans) != 0:
        for glycan, smiles in convert_generator(glycan_list=glycans):
            if returning:
                output.append((glycan, smiles))
            else:
                # ... and return them as intended
                print(glycan, smiles, file=output, sep=",")

    if glycan_generator is not None:
        for glycan, smiles in convert_generator(glycan_generator=glycan_generator):
            if returning:
                output.append((glycan, smiles))
            else:
                # ... and return them as intended
                print(glycan, smiles, file=output, sep=",")

    if returning:
        return output
    elif output_file is not None:
        output.close()


def convert_generator(glycan=None, glycan_list=None, glycan_file=None, glycan_generator=None, silent=True):
    """
    General user interaction interface to use this library.

    Args:
        glycan (str): Single glycan to be converted from IUPAC to SMILES
        glycan_list (List[str]): list of glycans to convert
        glycan_file (str): File to read the glycans from
        glycan_generator (generator): generator yielding iupac representation.
            Together with output_generator=True this does not create any lists
        silent (bool): Flag indicating to have no output from this method

    Returns:
        Nothing
    """

    glycans = preprocess_glycans(glycan, glycan_list, glycan_file)
    if len(glycans) == 0 and glycan_generator is None:
        if not silent:
            print("List of glycans is empty")
        return

    # Convert the glycans ...
    if len(glycans) != 0:
        for glycan in glycans:
            try:
                # ... by passing them to the glycan class to parse them and return them as intended
                if parsable_glycan(glycan):
                    yield glycan, Glycan(glycan).get_smiles()
                else:
                    print(f"Glycan {glycan} is not parsable due to unknown monomers.", file=sys.stderr)
                    yield glycan, ""

            # catch any exception at glycan level to not destroy the whole pipeline because of one mis-formed glycan
            except Exception as e:
                print(f"An exception occurred with {glycan}:", e.__class__, file=sys.stderr)
                print("Error message:", e.__str__(), file=sys.stderr)
                yield glycan, ""

    # Convert the glycans ...
    if glycan_generator is not None:
        for glycan in glycan_generator:
            try:
                # ... by passing them to the glycan class to parse them and return them as intended
                if parsable_glycan(glycan):
                    yield glycan, Glycan(glycan).get_smiles()
                else:
                    print(f"Glycan {glycan} is not parsable due to unknown monomers.", file=sys.stderr)
                    yield glycan, ""

            # catch any exception at glycan level to not destroy the whole pipeline because of one mis-formed glycan
            except Exception as e:
                print(f"An exception occurred with {glycan}:", e.__class__, file=sys.stderr)
                print("Error message:", e.__str__(), file=sys.stderr)
                yield glycan, ""



from glyles.glycans.factory import MonomerFactory
from glyles.glycans.utils import Config


class Monomer:
    """
    General interface to represent the sugar monomers in the tree.

    The resulting molecule must have some properties that are especially important for the correct implementation of
        the parser later on:
          - The C atoms have IDs 1 to 6 according to their C1-C6 naming in a glycan.
          - The O atom closing the ring has id 10, so that all other O-atoms have ID 10 higher than the C atom they are
            connected to.
          - The chirality is initial 0 and will only be set in the carbons with respect to the non-OH-group.
          - Additionally, all members of the ring of the molecule have an according flag set to true.

    Depending on the concrete implementation the actual ids of the atoms might vary, but the rest of the program will
    use the atom ids as specified here. So, an alternative naming must internally be converted to this specification.
    """

    def __init__(self, origin=None, **kwargs):
        """
        Initialize the Monomer with the provided arguments or from another monomer

        Args:
            origin (Monomer): Other monomer to use to initialize this object
            **kwargs: arguments to initialize monomer if object is None. Must include name, SMILES, and config
        """
        if origin is None:
            self._name = kwargs["name"]
            self._smiles = kwargs["smiles"]
            self._structure = kwargs.get("struct", None)
            self._config = kwargs.get("config", Config.UNDEF)
        else:
            self._name = origin.get_name()
            self._smiles = origin.get_smiles()
            self._structure = origin.get_structure()
            self._config = origin.get_config()

    def get_name(self):
        """
        Returns the name of this monomer as three-letter code (eventually longer for more fancy monosaccharides with
        more complex side chains).

        Returns:
            The name of this monomer
        """
        return self._name

    def get_smiles(self):
        """
        Returns the SMILES representation of this monomer. Attention: These methods return the SMILES that is used to
        initialize this monomer. This is different from the to_smiles method of this class that returns the SMILES
        string with added place-holders that is used for the generation of the SMILES representation of the complete
        glycan

        Returns:
            The SMILES string that was used for initialization of this monomer
        """
        return self._smiles

    def get_structure(self):
        """
        Returns the structure of this monomer. This might be Null in case the structure wasn't inferred from the
        SMILES so far

        Returns:
            Structure depending on the methods chosen to represent the structure of this monomer, might be Null
        """
        return self._structure

    def alpha(self):
        """
        Return this monosaccharide in its alpha conformation.

        Returns:
            Monomer in alpha conformation
        """
        return self.from_string("A_" + self._name)

    def beta(self):
        """
        Return this monosaccharide in its beta conformation.

        Returns:
            Monomer in beta conformation
        """
        return self.from_string("B_" + self._name)

    def undefined(self):
        """
        Return this monosaccharide in undefined conformation, the first carbon ring-atom will have unspecified.
        chirality.

        Returns:
            Monomer in undefined conformation
        """
        return self.from_string(self._name)

    def to_chirality(self, chirality):
        """
        Return this monomer in the queried chirality.
        Args:
            chirality (str): char representing the chiral conformation of the first carbon ring atom

        Returns:
            This monomer with the given (or not given) chirality at the first carbon ring atom
        """
        chirality = chirality.lower()
        if chirality == "a":
            return self.alpha()
        if chirality == "b":
            return self.beta()
        return self.undefined()

    def get_config(self):
        """
        The current conformation relative to the first carbon ring-atom, i.e. alpha, beta or unspecified.

        Returns:
            Config-Tag according to the conformation this monomer represents
        """
        return self._config

    def is_non_chiral(self):
        """
        Check if this monomer represents a non-chiral molecule

        Returns:
            boolean indicating chirality of this monomer
        """
        return self._config == Config.UNDEF

    def get_dummy_atoms(self):
        """
        Specify some dummy atoms that are used to mark oxygen atoms that will participate in bindings between glycans.

        Returns:
            Two lists:
                * one with atoms that are given as the "atom" argument to the mark-method to replace the oxygen atoms
                * the string representation of the atoms from above, i.e. how the atoms above will be represented in a
                  SMILES string
        """
        raise NotImplementedError

    def root_atom_id(self, binding_c_id):
        """
        Get ID of atom that will bind the parent monomer in the glycan. This ID will be given as root argument to
        the to_smiles method.

        Args:
            binding_c_id (int): Integer at which c-position this monomer binds its parent

        Returns:
            id of the atom that binds to the parent
        """
        raise NotImplementedError

    def mark(self, position, atom):
        """
        Mark the oxygen atom linked to the carbon atom at the given position ready to participate in the bounding.

        Args:
            position (int): id of the carbon atom whose oxygen atom will from the binding
            atom (object): atom to replace the binding oxygen with

        Returns:
            Nothing
        """
        raise NotImplementedError

    def to_smiles(self, root, ring_index):
        """
        Convert this monomer into a SMILES string representation.

        Args:
            root (int): index of the root atom
            ring_index (int): index of the rings in the atom

        Returns:
            SMILES string representation of this molecule
        """
        raise NotImplementedError

    def __get_structure(self):
        """
        Compute and save the structure of this glycan.

        Returns:
            Representation the structure of the glycan as a graph.
        """
        raise NotImplementedError

    @staticmethod
    def from_string(mono):
        """
        Get an instance of a glycan according to the provided string representation of the glycan.

        Args:
            mono (str): string representation of the glycan of interest

        Returns:
            Glycan according to the monosaccharide provided via mono
        """
        factory = MonomerFactory()
        args = factory[mono.upper()]
        return Monomer(**args)



from enum import Enum


class Config(Enum):
    """
    Configuration enum to represent if the monomer is in alpha, beta, or undefined configuration.
    """
    UNDEF = 0
    ALPHA = 1
    BETA = 2


class Enantiomer(Enum):
    """
    Configuration of the whole monomer regarding L and D forms in terms of enantiomerism.
    """
    D = 0
    L = 1



from enum import Enum

import networkx as nx

from glyles.glycans.monomer import Monomer


class NXMonomer(Monomer):
    """
    Different monosaccharides that can already be used in the library.
    They contain their name, one SMILES representation and a field to store the structure of the glycan once computed
    """

    class Chirality(Enum):
        """
        Representation of the chirality of certain atoms based on the Haworth notations of glycans.
        """
        UP = 1
        DOWN = 2
        NONE = 3

        @staticmethod
        def from_string(c):
            """
            Get a conformation flag from a string representation
            Args:
                c (str): name of the conformation in alpha or beta

            Returns:
                Conformation flag according to the char
            """
            if c.lower() == "a":
                return NXMonomer.Chirality.DOWN
            if c.lower() == "b":
                return NXMonomer.Chirality.UP
            return NXMonomer.Chirality.NONE

    class Atom(Enum):
        """
        Different types of atoms in the glycans.
        X, Y, Z are used to perform the concatenation of the glycans
        """
        N = "N"
        C = "C"
        O = "O"
        X = "X"
        Y = "Y"
        Z = "Z"

    def __init__(self, origin=None, **kwargs):
        """
        Initialize this class from parent class

        Args:
            origin (Monomer): other instance of Monomers to copy to this instance
            **kwargs: arguments to alternative initialize the monomer if origin is None
        """
        super(NXMonomer, self).__init__(origin, **kwargs)

    def get_dummy_atoms(self):
        """
        Specify some dummy atoms that are used to mark oxygen atoms that will participate in bindings between glycans.
        Here, the atoms will be replaced by instances of the atom enum that are used to define the type of the atom in
        the nodes of the networkx representation of the monomer molecules.

        Returns:
            Two lists:
                * one with atoms that are given as the "atom" argument to the mark-method to replace the oxygen atoms
                * the string representation of the atoms from above, i.e. how the atoms above will be represented in a
                  SMILES string
        """
        return [NXMonomer.Atom.X, NXMonomer.Atom.Y, NXMonomer.Atom.Z], ["X", "Y", "Z"]

    def root_atom_id(self, binding_c_id):
        """
        Get ID of atom that will bind the parent monomer in the glycan. This ID will be given as root argument to
        the to_smiles method.

        Args:
            binding_c_id (int): Integer at which c-position this monomer binds its parent

        Returns:
            id of the atom that binds to the parent, -1 if the root cannot be found
        """
        child_start = -1

        # find the ID of the oxygen atom to start the SMILES representation from
        for _, x in self.__get_structure().edges(binding_c_id):
            if self.__get_structure().nodes[x]["type"] == NXMonomer.Atom.O and 11 <= x <= 15:
                child_start = x
                break

        # if the start-point couldn't be found, raise an error
        if child_start == -1:
            raise ValueError("SMILES cannot computed from atom -1!")

        return child_start

    def mark(self, position, atom):
        """
        Mark the oxygen atom linked to the carbon atom at the given position ready to participate in the bounding.
        Marking here works based on replacing the oxygen-group bound to the carbon atom at the given position with the
        atom enum instance also provided in the arguments.

        Args:
            position (int): id of the carbon atom whose oxygen atom will from the binding
            atom (object): atom to replace the binding oxygen with

        Returns:
            Nothing
        """
        monomer = self.__get_structure()

        # and find that oxygen atom that will react with a hydrogen when connecting the glycans' monomers
        for _, x in monomer.edges(position):
            if monomer.nodes[x]["type"] == NXMonomer.Atom.O and 11 <= x <= 15:
                monomer.nodes[x]["type"] = atom
                break

    def to_smiles(self, root=10, ring_index=1):
        """
        Convert this monomer into a SMILES string representation.
        Use the implementation of the SMILES algorithm fitted to the needs of glycans.

        Args:
            root (int): index of the root atom
            ring_index (int): index of the rings in the atom

        Returns:
            SMILES string representation of this molecule
        """
        return SMILES().write(self.__get_structure(), root, ring_index)

    def __get_structure(self):
        """
        Compute and save the structure of this glycan. So far its hard coded on the 4 given types of glycans
        Chirality refers to the OH/CH2OH group of the chiral C atom

        Returns:
            networkx graph representing the structure of the glycan as a graph of its non-hydrogen atoms.
        """
        if self._structure is None:
            g = nx.Graph()
            g.add_nodes_from([
                (1, {"type": NXMonomer.Atom.C, "chiral": NXMonomer.Chirality.NONE, "ring": True}),
                (2, {"type": NXMonomer.Atom.C, "chiral": NXMonomer.Chirality.NONE, "ring": True}),
                (3, {"type": NXMonomer.Atom.C, "chiral": NXMonomer.Chirality.NONE, "ring": True}),
                (4, {"type": NXMonomer.Atom.C, "chiral": NXMonomer.Chirality.NONE, "ring": True}),
                (5, {"type": NXMonomer.Atom.C, "chiral": NXMonomer.Chirality.NONE, "ring": True}),
                (6, {"type": NXMonomer.Atom.C, "chiral": NXMonomer.Chirality.NONE, "ring": False}),
                (10, {"type": NXMonomer.Atom.O, "chiral": NXMonomer.Chirality.NONE, "ring": True}),
                (11, {"type": NXMonomer.Atom.O, "chiral": NXMonomer.Chirality.NONE, "ring": False}),
                (12, {"type": NXMonomer.Atom.O, "chiral": NXMonomer.Chirality.NONE, "ring": False}),
                (13, {"type": NXMonomer.Atom.O, "chiral": NXMonomer.Chirality.NONE, "ring": False}),
                (14, {"type": NXMonomer.Atom.O, "chiral": NXMonomer.Chirality.NONE, "ring": False}),
                (15, {"type": NXMonomer.Atom.O, "chiral": NXMonomer.Chirality.NONE, "ring": False}),
            ])
            if self._name == "Glc":
                g.add_edges_from([
                    (10, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 10),
                    (1, 11), (2, 12), (3, 13), (4, 14), (5, 6), (6, 15),
                ])
                nx.set_node_attributes(g, {2: {"chiral": NXMonomer.Chirality.DOWN},
                                           3: {"chiral": NXMonomer.Chirality.UP},
                                           4: {"chiral": NXMonomer.Chirality.DOWN},
                                           5: {"chiral": NXMonomer.Chirality.UP}})
            elif self._name == "Man":
                g.add_edges_from([
                    (10, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 10),
                    (1, 11), (2, 12), (3, 13), (4, 14), (5, 6), (6, 15)
                ])
                nx.set_node_attributes(g, {2: {"chiral": NXMonomer.Chirality.UP},
                                           3: {"chiral": NXMonomer.Chirality.UP},
                                           4: {"chiral": NXMonomer.Chirality.DOWN},
                                           5: {"chiral": NXMonomer.Chirality.UP}})
            elif self._name == "Gal":
                g.add_edges_from([
                    (10, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 10),
                    (1, 11), (2, 12), (3, 13), (4, 14), (5, 6), (6, 15),
                ])
                nx.set_node_attributes(g, {2: {"chiral": NXMonomer.Chirality.DOWN},
                                           3: {"chiral": NXMonomer.Chirality.UP},
                                           4: {"chiral": NXMonomer.Chirality.UP},
                                           5: {"chiral": NXMonomer.Chirality.UP}})
            elif self._name == "Tal":
                g.add_edges_from([
                    (10, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 10),
                    (1, 11), (2, 12), (3, 13), (4, 14), (5, 6), (6, 15),
                ])
                nx.set_node_attributes(g, {2: {"chiral": NXMonomer.Chirality.UP},
                                           3: {"chiral": NXMonomer.Chirality.UP},
                                           4: {"chiral": NXMonomer.Chirality.UP},
                                           5: {"chiral": NXMonomer.Chirality.UP}})

            if self._config == NXMonomer.Config.ALPHA:
                nx.set_node_attributes(g, {1: {"chiral": NXMonomer.Chirality.DOWN}})
            elif self._config == NXMonomer.Config.BETA:
                nx.set_node_attributes(g, {1: {"chiral": NXMonomer.Chirality.UP}})

            self._structure = g
        return self._structure

    @staticmethod
    def from_string(mono):
        """
        Convert the string representation of the monomer into an instance.

        Args:
            mono (str): string encoding of a monomer

        Returns:
            NXMonomer instance according to the string
        """
        return NXMonomer(origin=Monomer.from_string(mono))


class DFS:
    """
    Class computing a depth-first search tree from a graph.
    """

    def dfs_tree(self, g, source):
        """
        Compute a depth-first search tree on g starting in the given source node.

        Args:
            g (networkx.Graph): graph to compute the DFS tree for
            source (int): ID of the node to compute the DFS tree from

        Returns:
            A networkx.DiGraph representing the DFS tree of the input graph from the provided source node
        """
        dfst = nx.DiGraph()
        self.__dfs(g, dfst, None, source)

        return dfst

    def __dfs(self, g, dfst, parent, node):
        """
        Recursively perform the depth-first search in the tree. The decision in which order to explore the children is
        important to keep the SMILES-generating algorithm as simple as possible. Here, the search is oriented along the
        ring in the glycan and all side-chains are explored later.

        Args:
            g (networkx.Graph): the graph to compute the DFS tree for
            dfst (networkx.DiGraph): current state of the DFS tree
            parent (int): parent node the search comes from
            node (int): current node id

        Returns:
            Nothing
        """
        # add the current node to the tree
        dfst.add_node(node)

        # in case this node has a parent, add the according edge
        if parent is not None:
            dfst.add_edge(parent, node)

        # iterate over all children in a sorted way, sorted by the membership to the ring and the ID of the node
        children = sorted([c for _, c in list(g.edges(node)) if c != parent],
                          key=lambda c: (not g.nodes[c]["ring"], (c - node) % 10))
        for child in children:
            # if the node is new, explore it
            if child not in dfst.nodes:
                self.__dfs(g, dfst, node, child)

            # if the child already has been seen, and it's not the parent, we found the ring-closing atoms and mark them
            elif g.nodes[child]["ring"] and g.nodes[node]["ring"]:
                nx.set_node_attributes(dfst, {child: True, node: True}, "ring")


class SMILES:
    """
    Class to hold all methods related to converting a structure into SMILES representation and to generate a structure
    from a SMILES representation.
    """

    def __init__(self):
        self.ring_index = 1

    def write(self, g, source, ring_index=1):
        """
        Compute the SMILES string for the provided graph representing a molecule.
        This only works if called from an Oxygen atom, i.e. a non-chiral atom!

        Args:
            g (networkx.Graph): Graph representation of a molecule
            source (int): Atom-ID to start the SMILES from
            ring_index (int): index of the ring in this specific monosaccharide

        Returns:
            SMILES string of the molecule starting at the given atom
        """
        # compute the DFS tree for the molecule from that given root node
        dfs_tree = DFS().dfs_tree(g, source)

        # based on that DFS tree construct the SMILES
        return self.__construct(g, dfs_tree, source, ring_index)

    def read(self):
        """
        See the exact implementation guidelines in monomer.py description of the monomer description.

        Returns:
            Molecule represented in a graph
        """
        pass

    def __construct(self, g, tree, node, ring_index=1):
        """
        Recursively compute the SMILES from the structure and the according DFS tree. The current step of the search
        has to start in the provided node.

        Args:
            g (nx.Graph): graph of the molecule
            tree (nx.DiGraph): DFS tree on that molecule
            node (int): node id the construction is currently in

        Returns:
            SMILES representation of the subtree of the current node in tree
        """

        # check for chirality of the molecule. If it's not chiral, just add the current atom type to the output
        # IMPORTANT: This method assumes that the chiral C-atoms have a hydrogen atom opposed to the OH group
        if g.nodes[node]["chiral"] != NXMonomer.Chirality.NONE:
            output = "[C{value}H]".format(value=("@@" if g.nodes[node]["chiral"] == NXMonomer.Chirality.DOWN else "@"))
        else:
            output = g.nodes[node]["type"].value

        # check if the molecule is start or end of the ring, mark this, remember, there is only one ring in glycans
        if node in nx.get_node_attributes(tree, "ring"):
            output += str(ring_index)

        # recursively step down to the children
        children = list(tree.edges(node))
        if len(children) == 0:  # leaf
            return output
        else:  # children
            for c in [x[1] for x in children]:
                output += "({smiles})".format(smiles=self.__construct(g, tree, c))
            return output



from glyles.glycans.utils import Config, Enantiomer


class MonomerFactory:
    """
    Checked with http://www.cheminfo.org/flavor/malaria/Utilities/SMILES_generator___checker/index.html
    """

    __monomers = {
        # Fucose
        "FUC": {"name": "Fuc", "config": Config.UNDEF, "isomer": Enantiomer.L,
                "smiles": "C[C@@H]1OC(O)[C@@H](O)[C@H](O)[C@@H]1O"},
        "A_FUC": {"name": "Fuc", "config": Config.ALPHA, "isomer": Enantiomer.L,
                  "smiles": "C[C@@H]1O[C@@H](O)[C@@H](O)[C@H](O)[C@@H]1O"},
        "B_FUC": {"name": "Fuc", "config": Config.BETA, "isomer": Enantiomer.L,
                  "smiles": "C[C@@H]1O[C@H](O)[C@@H](O)[C@H](O)[C@@H]1O"},

        # Galactose
        "GAL": {"name": "Gal", "config": Config.UNDEF, "isomer": Enantiomer.D,
                "smiles": "OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@H]1O"},
        "A_GAL": {"name": "Gal", "config": Config.ALPHA, "isomer": Enantiomer.D,
                  "smiles": "OC[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@H]1O"},
        "B_GAL": {"name": "Gal", "config": Config.BETA, "isomer": Enantiomer.D,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@H]1O"},

        "GAL3S": {"name": "Gal3S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                  "smiles": "O=S(=O)([O-])O[C@@H]1[C@@H](O)C(O)O[C@H](CO)[C@@H]1O"},
        "A_GAL3S": {"name": "Gal3S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])O[C@@H]1[C@@H](O)[C@@H](O)O[C@H](CO)[C@@H]1O"},
        "B_GAL3S": {"name": "Gal3S", "config": Config.BETA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])O[C@@H]1[C@@H](O)[C@H](O)O[C@H](CO)[C@@H]1O"},

        "GAL3S4S": {"name": "Gal3S4S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])O[C@@H]1[C@@H](O)C(O)O[C@H](CO)[C@@H]1OS(=O)(=O)[O-]"},
        "A_GAL3S4S": {"name": "Gal3S4S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                      "smiles": "O=S(=O)([O-])O[C@@H]1[C@@H](O)[C@@H](O)O[C@H](CO)[C@@H]1OS(=O)(=O)[O-]"},
        "B_GAL3S4S": {"name": "Gal3S4S", "config": Config.BETA, "isomer": Enantiomer.D,
                      "smiles": "O=S(=O)([O-])O[C@@H]1[C@@H](O)[C@H](O)O[C@H](CO)[C@@H]1OS(=O)(=O)[O-]"},

        "GAL3S6S": {"name": "Gal3S6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])OC[C@H]1OC(O)[C@H](O)[C@@H](OS(=O)(=O)[O-])[C@H]1O"},
        "A_GAL3S6S": {"name": "Gal3S6S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                      "smiles": "O=S(=O)([O-])OC[C@H]1O[C@H](O)[C@H](O)[C@@H](OS(=O)(=O)[O-])[C@H]1O"},
        "B_GAL3S6S": {"name": "Gal3S6S", "config": Config.BETA, "isomer": Enantiomer.D,
                      "smiles": "O=S(=O)([O-])OC[C@H]1O[C@@H](O)[C@H](O)[C@@H](OS(=O)(=O)[O-])[C@H]1O"},

        "GAL4S": {"name": "Gal4S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                  "smiles": "O=S(=O)([O-])O[C@@H]1[C@H](O)[C@@H](O)C(O)O[C@@H]1CO"},
        "A_GAL4S": {"name": "Gal4S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])O[C@@H]1[C@H](O)[C@@H](O)[C@@H](O)O[C@@H]1CO"},
        "B_GAL4S": {"name": "Gal4S", "config": Config.BETA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])O[C@@H]1[C@H](O)[C@@H](O)[C@H](O)O[C@@H]1CO"},

        "GAL4S6S": {"name": "Gal4S6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@H]1OS(=O)(=O)[O-]"},
        "A_GAL4S6S": {"name": "Gal4S6S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                      "smiles": "O=S(=O)([O-])OC[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@H]1OS(=O)(=O)[O-]"},
        "B_GAL4S6S": {"name": "Gal4S6S", "config": Config.BETA, "isomer": Enantiomer.D,
                      "smiles": "O=S(=O)([O-])OC[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@H]1OS(=O)(=O)[O-]"},

        "GAL6S": {"name": "Gal6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                  "smiles": "O=S(=O)([O-])OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@H]1O"},
        "A_GAL6S": {"name": "Gal6S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])OC[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@H]1O"},
        "B_GAL6S": {"name": "Gal6S", "config": Config.BETA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])OC[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@H]1O"},

        "GALNAC": {"name": "GalNAc", "config": Config.UNDEF, "isomer": Enantiomer.D,
                   "smiles": "CC(=O)N[C@H]1C(O)O[C@H](CO)[C@H](O)[C@@H]1O"},
        "A_GALNAC": {"name": "GalNAc", "config": Config.ALPHA, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1[C@@H](O)O[C@H](CO)[C@H](O)[C@@H]1O"},
        "B_GALNAC": {"name": "GalNAc", "config": Config.BETA, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1[C@H](O)O[C@H](CO)[C@H](O)[C@@H]1O"},

        "GALNAC3S": {"name": "GalNAc3S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1C(O)O[C@H](CO)[C@H](O)[C@@H]1OS(=O)(=O)[O-]"},
        "A_GALNAC3S": {"name": "GalNAc3S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@@H](O)O[C@H](CO)[C@H](O)[C@@H]1OS(=O)(=O)[O-]"},
        "B_GALNAC3S": {"name": "GalNAc3S", "config": Config.BETA, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@H](O)O[C@H](CO)[C@H](O)[C@@H]1OS(=O)(=O)[O-]"},

        "GALNAC4S": {"name": "GalNAc4S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1C(O)O[C@H](CO)[C@H](OS(=O)(=O)O)[C@@H]1O"},
        "A_GALNAC4S": {"name": "GalNAc4S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@@H](O)O[C@H](CO)[C@H](OS(=O)(=O)O)[C@@H]1O"},
        "B_GALNAC4S": {"name": "GalNAc4S", "config": Config.BETA, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@H](O)O[C@H](CO)[C@H](OS(=O)(=O)O)[C@@H]1O"},

        "GALNAC4S6S": {"name": "GalNAc4S6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1C(O)O[C@H](COS(=O)(=O)[O-])[C@H](OS(=O)(=O)[O-])[C@@H]1O"},
        "A_GALNAC4S6S": {"name": "GalNAc4S6S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                         "smiles": "CC(=O)N[C@H]1[C@@H](O)O[C@H](COS(=O)(=O)[O-])[C@H](OS(=O)(=O)[O-])[C@@H]1O"},
        "B_GALNAC4S6S": {"name": "GalNAc4S6S", "config": Config.BETA, "isomer": Enantiomer.D,
                         "smiles": "CC(=O)N[C@H]1[C@H](O)O[C@H](COS(=O)(=O)[O-])[C@H](OS(=O)(=O)[O-])[C@@H]1O"},

        "GALNAC6S": {"name": "GalNAc6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1C(O)O[C@H](COS(=O)(=O)[O-])[C@H](O)[C@@H]1O"},
        "A_GALNAC6S": {"name": "GalNAc6S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@@H](O)O[C@H](COS(=O)(=O)[O-])[C@H](O)[C@@H]1O"},
        "B_GALNAC6S": {"name": "GalNAc6S", "config": Config.BETA, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@H](O)O[C@H](COS(=O)(=O)[O-])[C@H](O)[C@@H]1O"},

        # Glucose
        "GLC": {"name": "Glc", "config": Config.UNDEF, "isomer": Enantiomer.D,
                "smiles": "OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "A_GLC": {"name": "Glc", "config": Config.ALPHA, "isomer": Enantiomer.D,
                  "smiles": "OC[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "B_GLC": {"name": "Glc", "config": Config.BETA, "isomer": Enantiomer.D,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},

        "GLC4S": {"name": "Gal4S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                  "smiles": "O=S(=O)([O-])O[C@H]1[C@H](O)[C@@H](O)C(O)O[C@@H]1CO"},
        "A_GLC4S": {"name": "Gal4S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])O[C@H]1[C@H](O)[C@@H](O)[C@@H](O)O[C@@H]1CO"},
        "B_GLC4S": {"name": "Gal4S", "config": Config.BETA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])O[C@H]1[C@H](O)[C@@H](O)[C@H](O)O[C@@H]1CO"},

        "GLC6S": {"name": "Glc6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                  "smiles": "O=S(=O)([O-])OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "A_GLC6S": {"name": "Glc6S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])OC[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "B_GLC6S": {"name": "Glc6S", "config": Config.BETA, "isomer": Enantiomer.D,
                    "smiles": "O=S(=O)([O-])OC[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},

        "GLCA": {"name": "GlcA", "config": Config.UNDEF, "isomer": Enantiomer.D,
                 "smiles": "O=C(O)[C@H]1OC(O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "A_GLCA": {"name": "GlcA", "config": Config.ALPHA, "isomer": Enantiomer.D,
                   "smiles": "O=C(O)[C@H]1O[C@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},
        "B_GLCA": {"name": "GlcA", "config": Config.BETA, "isomer": Enantiomer.D,
                   "smiles": "O=C(O)[C@H]1O[C@@H](O)[C@H](O)[C@@H](O)[C@@H]1O"},

        "GLCN": {"name": "GlcN", "config": Config.UNDEF, "isomer": Enantiomer.D,
                 "smiles": "N[C@H]1C(O)O[C@H](CO)[C@@H](O)[C@@H]1O"},
        "A_GLCN": {"name": "GlcN", "config": Config.ALPHA, "isomer": Enantiomer.D,
                   "smiles": "N[C@H]1[C@@H](O)O[C@H](CO)[C@@H](O)[C@@H]1O"},
        "B_GLCN": {"name": "GlcN", "config": Config.BETA, "isomer": Enantiomer.D,
                   "smiles": "N[C@H]1[C@H](O)O[C@H](CO)[C@@H](O)[C@@H]1O"},

        "GLCNAC": {"name": "GlcNAc", "config": Config.UNDEF, "isomer": Enantiomer.D,
                   "smiles": "CC(=O)N[C@H]1C(O)O[C@H](CO)[C@@H](O)[C@@H]1O"},
        "A_GLCNAC": {"name": "GlcNAc", "config": Config.UNDEF, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1[C@@H](O)O[C@H](CO)[C@@H](O)[C@@H]1O"},
        "B_GLCNAC": {"name": "GlcNAc", "config": Config.UNDEF, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1[C@H](O)O[C@H](CO)[C@@H](O)[C@@H]1O"},

        "GLCNAC3S": {"name": "GalNAc3S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1C(O)O[C@H](CO)[C@@H](O)[C@@H]1OS(=O)(=O)[O-]"},
        "A_GLCNAC3S": {"name": "GalNAc3S", "config": Config.ALPHA, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@@H](O)O[C@H](CO)[C@@H](O)[C@@H]1OS(=O)(=O)[O-]"},
        "B_GLCNAC3S": {"name": "GalNAc3S", "config": Config.BETA, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@H](O)O[C@H](CO)[C@@H](O)[C@@H]1OS(=O)(=O)[O-]"},

        "GLCNAC6S": {"name": "GlcNAc6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                     "smiles": "CC(=O)N[C@H]1C(O)O[C@H](COS(=O)(=O)[O-])[C@@H](O)[C@@H]1O"},
        "A_GLCNAC6S": {"name": "GlcNAc6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@@H](O)O[C@H](COS(=O)(=O)[O-])[C@@H](O)[C@@H]1O"},
        "B_GLCNAC6S": {"name": "GlcNAc6S", "config": Config.UNDEF, "isomer": Enantiomer.D,
                       "smiles": "CC(=O)N[C@H]1[C@H](O)O[C@H](COS(=O)(=O)[O-])[C@@H](O)[C@@H]1O"},

        # Mannose
        "MAN": {"name": "Man", "config": Config.UNDEF, "isomer": Enantiomer.D,
                "smiles": "OC[C@H]1OC(O)[C@@H](O)[C@@H](O)[C@@H]1O"},
        "A_MAN": {"name": "Man", "config": Config.ALPHA, "isomer": Enantiomer.D,
                  "smiles": "OC[C@H]1O[C@H](O)[C@@H](O)[C@@H](O)[C@@H]1O"},
        "B_MAN": {"name": "Man", "config": Config.BETA, "isomer": Enantiomer.D,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@@H](O)[C@@H](O)[C@@H]1O"},

        # Talose
        "TAL": {"name": "Tal", "config": Config.UNDEF, "isomer": Enantiomer.D,
                "smiles": "OC[C@H]1OC(O)[C@@H](O)[C@@H](O)[C@H]1O"},
        "A_TAL": {"name": "Tal", "config": Config.ALPHA, "isomer": Enantiomer.D,
                  "smiles": "OC[C@H]1O[C@H](O)[C@@H](O)[C@@H](O)[C@H]1O"},
        "B_TAL": {"name": "Tal", "config": Config.BETA, "isomer": Enantiomer.D,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@@H](O)[C@@H](O)[C@H]1O"},
    }

    '''
    # Keto - Deoxy - Nonulonic acid
    "KDN": {"name": "Kdn", "config": Config.UNDEF, "smiles": "O=C(O)C1(O)C[C@H](O)[C@@H](O)C([C@H](O)[C@H](O)CO)O1"},
    "A_KDN": {"name": "Kdn", "config": Config.UNDEF, "smiles": "O=C(O)[C@@]1(O)C[C@H](O)[C@@H](O)C([C@H](O)[C@H](O)CO)O1"},
    "B_KDN": {"name": "Kdn", "config": Config.UNDEF, "smiles": "O=C(O)[C@]1(O)C[C@H](O)[C@@H](O)C([C@H](O)[C@H](O)CO)O1"},
    
    # Neuraminic acid
    "NEU5AC": {"name": "Neu5Ac", "config": Config.UNDEF, "smiles": "CC(=O)N[C@@H]1[C@@H](O)CC(O)(C(=O)O)OC1[C@H](O)[C@H](O)CO"},
    "A_NEU5AC": {"name": "Neu5Ac", "config": Config.ALPHA, "smiles": "CC(=O)N[C@@H]1[C@@H](O)C[C@@](O)(C(=O)O)OC1[C@H](O)[C@H](O)CO"},
    "B_NEU5AC": {"name": "Neu5Ac", "config": Config.BETA, "smiles": "CC(=O)N[C@@H]1[C@@H](O)C[C@](O)(C(=O)O)OC1[C@H](O)[C@H](O)CO"},

    "NEU5GC": {"name": "Neu5Ac", "config": Config.UNDEF, "smiles": "O=C(CO)N[C@@H]1[C@@H](O)CC(O)(C(=O)O)OC1[C@H](O)[C@H](O)CO"},
    "A_NEU5GC": {"name": "Neu5Ac", "config": Config.ALPHA, "smiles": "O=C(CO)N[C@@H]1[C@@H](O)C[C@](O)(C(=O)O)OC1[C@H](O)[C@H](O)CO"},
    "B_NEU5GC": {"name": "Neu5Ac", "config": Config.BETA, "smiles": "O=C(CO)N[C@@H]1[C@@H](O)C[C@@](O)(C(=O)O)OC1[C@H](O)[C@H](O)CO"},
    '''

    def __contains__(self, item):
        return item.upper() in MonomerFactory.__monomers.keys()

    @staticmethod
    def keys():
        return MonomerFactory.__monomers.keys()

    @staticmethod
    def monomers():
        return list(set(x.split("_")[-1] for x in MonomerFactory.__monomers.keys()))

    @staticmethod
    def __getitem__(item):
        return MonomerFactory.__monomers[item.upper()]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glyles', 'glyles.glycans', 'glyles.grammar']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime>=4.9.3,<5.0.0',
 'networkx>=2.6.3,<3.0.0',
 'numpy>=1.21.4,<2.0.0',
 'pydot>=1.4.2,<2.0.0',
 'rdkit-pypi>=2021.9.2,<2022.0.0']

setup_kwargs = {
    'name': 'glyles',
    'version': '0.0.2',
    'description': 'A tool to convert IUPAC representation of glycans into SMILES strings',
    'long_description': None,
    'author': 'Roman Joeres',
    'author_email': 'Roman.Joeres@helmholtz-hzi.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import numpy as np
from rdkit.Chem import MolFromSmiles, MolToSmiles, GetAdjacencyMatrix

from glyles.glycans.monomer import Monomer


class RDKitMonomer(Monomer):
    def __init__(self, origin=None, **kwargs):
        """
        Initialize the monomer using the super method. Additionally, some fields are initialized to describe the
        structure of the monomer according to the specification of the monomer-parent class

        Args:
            origin (Monomer): Other monomer to use to initialize this object
            **kwargs: arguments to initialize monomer if object is None. Must include name, SMILES, and config
        """
        super(RDKitMonomer, self).__init__(origin, **kwargs)
        if isinstance(origin, RDKitMonomer):
            self._adjacency = origin.get_adjacency()
            self._ring_info = origin.get_ring_info()
            self._x = origin.get_features()
        else:
            self._adjacency = None
            self._ring_info = None
            self._x = None
            self.__get_structure()

    def get_adjacency(self):
        """
        Get the adjacency matrix of the atoms in this monomer.

        Returns:
            Adjacency matrix of all non-hydrogen atoms in this monomer
        """
        return self._adjacency

    def get_ring_info(self):
        """
        Get information of all atom-ids in the rings of this monomer.

        Returns:
            Tuple of tuples with the atom-ids from rdkit in the monomer
        """
        return self._ring_info

    def get_features(self):
        """
        Get a feature matrix from this monomer. The features contain information about atom type, ids according to the
        specification in monomer.py, and ring memberships.

        Returns:
            A numpy array of shape Nx3 containing the extracted features for all atoms in this molecule
        """
        return self._x

    def get_dummy_atoms(self):
        """
        Specify some dummy atoms that are used to mark oxygen atoms that will participate in bindings between glycans.
        Here, the atoms will be replaced by instances of the atom enum that are used to define the type of the atom in
        the nodes of the networkx representation of the monomer molecules.

        Returns:
            Two lists:
                * one with atoms that are given as the "atom" argument to the mark-method to replace the oxygen atoms
                * the string representation of the atoms from above, i.e. how the atoms above will be represented in a
                  SMILES string
        """
        return [34, 52, 84], ["[SeH]", "[TeH]", "[PoH]"]

    def root_atom_id(self, binding_c_id):
        """
        Get ID of atom that will bind the parent monomer in the glycan. This ID will be given as root argument to
        the to_smiles method.

        Args:
            binding_c_id (int): Integer at which c-position this monomer binds its parent

        Returns:
            id of the atom that binds to the parent, -1 if the root cannot be found
        """
        return self.__find_oxygen(binding_c_id)

    def mark(self, position, atom):
        """
        Mark the oxygen atom linked to the carbon atom at the given position ready to participate in the bounding.
        Marking here works based on replacing the oxygen-group bound to the carbon atom at the given position with the
        atom enum instance also provided in the arguments.

        Args:
            position (int): id of the carbon atom whose oxygen atom will from the binding
            atom (object): atom to replace the binding oxygen with

        Returns:
            Nothing
        """
        idx = self.__find_oxygen(position)
        self.__get_structure().GetAtomWithIdx(idx).SetAtomicNum(atom)
        self._x[idx, 0] = atom

    def to_smiles(self, root, ring_index):
        """
        Convert this monomer into a SMILES string representation.
        Use the implementation of the SMILES algorithm fitted to the needs of glycans.

        Args:
            root (int): index of the root atom
            ring_index (int): index of the rings in the atom

        Returns:
            SMILES string representation of this molecule
        """
        smiles = MolToSmiles(self.__get_structure(), rootedAtAtom=root)
        return "".join([(str(int(c) + ring_index) if c.isdigit() else c) for c in smiles])

    def __get_structure(self):
        """
        Compute and save the structure of this glycan.

        Returns:
            rdkit molecule representing the structure of the glycan as a graph of its non-hydrogen atoms.
        """
        if self._structure is None:
            # read the structure from the SMILES string
            self._structure = MolFromSmiles(self._smiles)

            # extract some further information from the molecule to not operate always on the molecule
            self._adjacency = GetAdjacencyMatrix(self._structure)
            self._ring_info = self._structure.GetRingInfo().AtomRings()
            self._x = np.zeros((self._adjacency.shape[0], 3))

            # extract some information form the molecule
            for i in range(self._adjacency.shape[0]):
                atom = self._structure.GetAtomWithIdx(i)

                # store the atom type
                self._x[i, 0] = atom.GetAtomicNum()

                # if the atom is part of any ring, store the number of that ring
                for r in range(len(self._ring_info)):
                    if i in self._ring_info[r]:
                        self._x[i, 2] = r + 1

                # identify the oxygen atom in the main ring and set its id to 10
                if self._x[i, 2] == 1 and self._x[i, 0] == 8:
                    self._x[i, 1] = 10

            # check in which order the ring has to be traversed to find C1 as first carbon atom behind the ring-oxygen
            main_ring = list(self._ring_info[0]) + list(self._ring_info[0])
            if not self.__clockwise():
                main_ring = list(reversed(main_ring))

            # TODO: This only works if the C1 atom is part of the ring! (i.e. not for fructose)
            # for all carbon atoms in the main ring, set its id according to the distance to the oxygen atom
            # iterate clockwise, i.e. from ring-O to C1/C2 and so on ...
            carbon_count = 1
            # Insert code for c1-atoms outside the ring here
            o_index = main_ring.index(np.argwhere((self._x[:, 0] == 8) & (self._x[:, 2] == 1)))
            for r in self._ring_info[0]:
                if self._x[r, 0] != 8:
                    self._x[r, 1] = carbon_count + main_ring.index(r, o_index) - o_index - 1

            # assign ids to the remaining carbon atoms
            highest_index = np.max(self._x[:, 1][np.argwhere(self._x[:, 0] == 6)])
            # c6_candidates = np.argwhere((self._x[:, 1] == 6) & (self._x[:, 2] == 0))

            c5 = np.argwhere((self._x[:, 1] == 5).squeeze())
            c6 = np.argwhere(((self._x[:, 0] == 6) & (self._x[:, 2] == 0) & (self._adjacency[c5, :] == 1)).squeeze())
            self._x[c6, 1] = highest_index + 1
        return self._structure

    def __clockwise(self):
        """
        Check if the main ring in the monomer is iterated in clockwise manner. RDKit assigns the atom_ids in the order
        it is read in during parsing the monomer SMILES strings.

        Returns:
            True if the main ring is traversed clockwise when iterating over all ring member in order of their RDKit-ID
        """
        position = np.argwhere(self._x[:, 1] == 10)

        # then find the candidates. There should be exactly one element in the resulting array
        candidates = np.argwhere((self._adjacency[position, :] == 1).squeeze())
        if len(candidates) != 2:
            raise ValueError("Parsing exception")

        # check if the candidate with the lower id fulfills the criteria to be the C1 atom
        first_c_cons = np.argwhere((self._adjacency[candidates[0], :] == 1).squeeze())
        if np.sum(self._x[first_c_cons, 0] == 6) == 1:
            o_index = self._ring_info[0].index(np.argwhere((self._x[:, 0] == 8) & (self._x[:, 2] == 1)))
            c_index = self._ring_info[0].index(candidates[0])
            if o_index < c_index:
                return True
        return False

    def __find_oxygen(self, binding_c_id):
        """
        Find the oxygen atom that binds to the carbon atom with the provided id. The returned id may not refer to an
        oxygen atom in the ring of the monomer as this cannot bind anything.

        Args:
            binding_c_id (int): id of the carbon atom that participates in a binding, and we need to find the oxygen
            from

        Returns:
            id referring to the oxygen binding the provided carbon atom and may participate in a glycan-binding.
        """
        # first find the rdkit id of the carbon atom that should bind to something
        position = np.argwhere(self._x[:, 1] == binding_c_id).squeeze()

        # then find the candidates. There should be exactly one element in the resulting array
        candidates = np.argwhere((self._adjacency[position, :] == 1) &
                                 (self._x[:, 0] == 8) & (self._x[:, 2] != 1)).squeeze()
        return int(candidates)

    @staticmethod
    def from_string(mono):
        """
        Convert the string representation of the monomer into an instance.

        Args:
            mono (str): string encoding of a monomer

        Returns:
            RDKit instance according to the string
        """
        tmp = Monomer.from_string(mono)
        return RDKitMonomer(origin=tmp)



# Generated from Glycan.g4 by ANTLR 4.9
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\t")
        buf.write("\'\4\2\t\2\4\3\t\3\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2")
        buf.write("\3\2\3\2\3\2\5\2\23\n\2\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3")
        buf.write("\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\5\3%\n\3\3\3\2\2\4")
        buf.write("\2\4\2\2\2*\2\22\3\2\2\2\4$\3\2\2\2\6\7\5\4\3\2\7\b\7")
        buf.write("\6\2\2\b\t\7\3\2\2\t\n\7\b\2\2\n\23\3\2\2\2\13\f\5\4\3")
        buf.write("\2\f\r\7\6\2\2\r\23\3\2\2\2\16\17\7\6\2\2\17\20\7\3\2")
        buf.write("\2\20\23\7\b\2\2\21\23\7\6\2\2\22\6\3\2\2\2\22\13\3\2")
        buf.write("\2\2\22\16\3\2\2\2\22\21\3\2\2\2\23\3\3\2\2\2\24\25\7")
        buf.write("\6\2\2\25%\7\7\2\2\26\27\7\6\2\2\27\30\7\7\2\2\30%\5\4")
        buf.write("\3\2\31\32\7\4\2\2\32\33\5\4\3\2\33\34\7\5\2\2\34%\3\2")
        buf.write("\2\2\35\36\7\6\2\2\36\37\7\7\2\2\37 \7\4\2\2 !\5\4\3\2")
        buf.write("!\"\7\5\2\2\"#\5\4\3\2#%\3\2\2\2$\24\3\2\2\2$\26\3\2\2")
        buf.write("\2$\31\3\2\2\2$\35\3\2\2\2%\5\3\2\2\2\4\22$")
        return buf.getvalue()


class GlycanParser ( Parser ):

    grammarFileName = "Glycan.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "' '", "'['", "']'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "SAC", "CON", "TYPE", "NUM" ]

    RULE_start = 0
    RULE_branch = 1

    ruleNames =  [ "start", "branch" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    SAC=4
    CON=5
    TYPE=6
    NUM=7

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.9")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class StartContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def branch(self):
            return self.getTypedRuleContext(GlycanParser.BranchContext,0)


        def SAC(self):
            return self.getToken(GlycanParser.SAC, 0)

        def TYPE(self):
            return self.getToken(GlycanParser.TYPE, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_start

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStart" ):
                listener.enterStart(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStart" ):
                listener.exitStart(self)




    def start(self):

        localctx = GlycanParser.StartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_start)
        try:
            self.state = 16
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 4
                self.branch()
                self.state = 5
                self.match(GlycanParser.SAC)
                self.state = 6
                self.match(GlycanParser.T__0)
                self.state = 7
                self.match(GlycanParser.TYPE)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 9
                self.branch()
                self.state = 10
                self.match(GlycanParser.SAC)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 12
                self.match(GlycanParser.SAC)
                self.state = 13
                self.match(GlycanParser.T__0)
                self.state = 14
                self.match(GlycanParser.TYPE)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 15
                self.match(GlycanParser.SAC)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BranchContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def SAC(self):
            return self.getToken(GlycanParser.SAC, 0)

        def CON(self):
            return self.getToken(GlycanParser.CON, 0)

        def branch(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.BranchContext)
            else:
                return self.getTypedRuleContext(GlycanParser.BranchContext,i)


        def getRuleIndex(self):
            return GlycanParser.RULE_branch

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBranch" ):
                listener.enterBranch(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBranch" ):
                listener.exitBranch(self)




    def branch(self):

        localctx = GlycanParser.BranchContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_branch)
        try:
            self.state = 34
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 18
                self.match(GlycanParser.SAC)
                self.state = 19
                self.match(GlycanParser.CON)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 20
                self.match(GlycanParser.SAC)
                self.state = 21
                self.match(GlycanParser.CON)
                self.state = 22
                self.branch()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 23
                self.match(GlycanParser.T__1)
                self.state = 24
                self.branch()
                self.state = 25
                self.match(GlycanParser.T__2)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 27
                self.match(GlycanParser.SAC)
                self.state = 28
                self.match(GlycanParser.CON)
                self.state = 29
                self.match(GlycanParser.T__1)
                self.state = 30
                self.branch()
                self.state = 31
                self.match(GlycanParser.T__2)
                self.state = 32
                self.branch()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx





